# This file is part of the project mb-netmgmt
#
# (C) 2022 Deutsche Telekom AG
#
# Deutsche Telekom AG and all other contributors / copyright
# owners license this file to you under the terms of the GPL-2.0:
#
# mb-netmgmt is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as published by
# the Free Software Foundation.
#
# mb-netmgmt is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with mb-netmgmt. If not, see <https://www.gnu.org/licenses/

"""Network Management Protocols for Mountebank"""
__version__ = "0.0.21"

import os
import subprocess
import time
from contextlib import contextmanager

import requests
import yaml


@contextmanager
def mb(imposters, loglevel="info"):
    with subprocess.Popen(
        ["mb", "--loglevel", loglevel], cwd=os.path.dirname(__file__)
    ) as mb:
        try:
            put_imposters("localhost", imposters)
            yield mb
        finally:
            mb.terminate()


def put_imposters(host, imposters):
    while True:
        try:
            response = requests.put(
                f"http://{host}:2525/imposters",
                json={"imposters": imposters},
            )
            break
        except requests.ConnectionError:
            time.sleep(1)
    try:
        response.raise_for_status()
    except requests.HTTPError as e:
        raise RuntimeError(e.response.json()["errors"])


def dump_imposters(host, name):
    response = requests.get(
        f"http://{host}:2525/imposters",
        {"replayable": True, "removeProxies": True},
    )
    yaml.safe_dump(response.json(), open(f"{name}.yaml", "w"))


def load_imposters(host, name):
    put_imposters(host, read_imposters(name))


def read_imposters(name):
    return yaml.safe_load(open(f"{name}.yaml").read())["imposters"]



# This file is part of the project mb-netmgmt
#
# (C) 2022 Deutsche Telekom AG
#
# Deutsche Telekom AG and all other contributors / copyright
# owners license this file to you under the terms of the GPL-2.0:
#
# mb-netmgmt is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as published by
# the Free Software Foundation.
#
# mb-netmgmt is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with mb-netmgmt. If not, see <https://www.gnu.org/licenses/

from mb_netmgmt.ssh import Server, Handler as SshHandler


class Handler(SshHandler):
    message_terminators = [b"]]>]]>", b"\n##\n"]
    default_port = 830

    def open_upstream(self):
        SshHandler.open_upstream(self)
        self.upstream_channel.invoke_subsystem("netconf")

    def handle_prompt(self):
        pass



import os
from threading import Thread

import ncclient.manager
import paramiko
import pytest
from mb_netmgmt import mb, netconf, ssh
from mb_netmgmt.__main__ import create_server

port = 8080
prompt = b"prompt"


@pytest.mark.parametrize("protocol", ["http", "snmp", "telnet", "netconf"])
def test_create_imposter(protocol):
    with mb([{"protocol": protocol, "port": port}]):
        pass


def test_ssh():
    with mb([{"protocol": "ssh", "port": port, "stubs": [prompt_stub()]}], "debug"):
        client = connect_ssh()
        chan = client.invoke_shell()
        out = chan.recv(1024)
        assert out == prompt


def test_ssh_proxy():
    with mb(
        [
            {
                "protocol": "ssh",
                "port": port,
                "stubs": [
                    prompt_stub(),
                    {
                        "responses": [
                            {
                                "proxy": {
                                    "to": f"ssh://{os.environ['NETCONF_USERNAME']}:{os.environ['NETCONF_PASSWORD']}@localhost"
                                }
                            },
                        ]
                    },
                ],
            }
        ]
    ):
        client = connect_ssh()
        chan = client.invoke_shell()
        out = chan.recv(1024)
        assert out == prompt


def prompt_stub():
    return {"responses": [{"is": {"response": prompt}}]}


def test_create_ssh_server():
    server = create_server(ssh, port, None)
    Thread(target=server.serve_forever).start()
    connect_ssh().close()
    server.shutdown()


def connect_ssh():
    client = paramiko.SSHClient()
    client.set_missing_host_key_policy(paramiko.AutoAddPolicy)
    client.connect("localhost", port)
    return client


def test_create_netconf_server():
    port = 8830
    netconf.Handler.open_upstream = lambda handler: None
    netconf.Handler.post_request = mock_post_request
    server = create_server(netconf, port, None)
    Thread(target=server.serve_forever).start()
    with ncclient.manager.connect(
        host="localhost", port=port, password="", hostkey_verify=False
    ) as m:
        m.get_config("running")
    ssh.stopped = True
    server.shutdown()


def mock_post_request(handler, request):
    command = request["command"]
    if "<nc:hello" in command:
        return {"response": {"response": "<hello/>]]>]]>"}}
    return {
        "response": {
            "response": '<rpc-reply xmlns="urn:ietf:params:xml:ns:netconf:base:1.0" message-id=""/>]]>]]>'
        }
    }



# This file is part of the project mb-netmgmt
#
# (C) 2022 Deutsche Telekom AG
#
# Deutsche Telekom AG and all other contributors / copyright
# owners license this file to you under the terms of the GPL-2.0:
#
# mb-netmgmt is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as published by
# the Free Software Foundation.
#
# mb-netmgmt is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with mb-netmgmt. If not, see <https://www.gnu.org/licenses/

import re
import telnetlib
import time
from socketserver import StreamRequestHandler, TCPServer

from mb_netmgmt.__main__ import Protocol

Server = TCPServer


class Handler(StreamRequestHandler, Protocol):
    def handle(self):
        self.callback_url = self.server.callback_url
        self.handle_username_prompt()
        self.username = extract_username(self.rfile.readline())
        self.handle_username()
        self.read_password()
        self.command_prompt = b"#"
        try:
            self.telnet.write(self.password)
            self.command_prompt = self.telnet.read_until(b"#").split(b"\n")[-1]
        except AttributeError:
            pass
        self.wfile.write(self.command_prompt)
        request = None
        while request != b"exit\r\n":
            request, request_id = self.read_request()
            if not request:
                return
            self.handle_command(request, request_id)

    def handle_command(self, command, request_id):
        return self.handle_request({"command": command.decode()}, request_id)

    def send_upstream(self, request, request_id):
        self.telnet.write(request["command"].encode())

    def read_request(self):
        return self.rfile.readline(), None

    def respond(self, response, request_id):
        self.wfile.write(response["response"].encode())

    def read_proxy_response(self):
        return {"response": self.telnet.read_until(self.command_prompt).decode()}

    def handle_username_prompt(self):
        username_prompt = b"Username: "
        to = self.get_to()
        if to:
            self.telnet = telnetlib.Telnet(to.hostname)
            username_prompt = self.telnet.read_until(username_prompt)
        self.wfile.write(username_prompt)

    def read_password(self):
        self.password = self.rfile.readline()
        while not self.password:
            time.sleep(1)
            self.password = self.rfile.readline()

    def handle_username(self):
        self.command_prompt = b"Password: "
        self.handle_request({"command": self.username.decode()}, None)


def extract_username(byte_string):
    return re.match(b".*?([-0-9a-zA-Z_].*\n)", byte_string).group(1)
